/*
 * This file is part of the IPCop Firewall.
 *
 * IPCop is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * IPCop is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with IPCop; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * Copyright (C) 2003-06-25 Tim Butterfield <timbutterfield@mindspring.com>
 *
 * $Id: ipcoprscfg.c,v 1.2.2.12 2007/04/30 20:21:36 gespinasse Exp $
 *
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <grp.h>
#include <dirent.h>
#include "setuid.h"

/* defines for config create/restore return status*/
#define ERR_ANY      1   // unspecified error
#define ERR_KEY      2   // error creating key file
#define ERR_TAR      3   // error creating .tar
#define ERR_GZ       4   // error creating .tar.gz
#define ERR_ENCRYPT  5   // error creating .dat
#define ERR_DECRYPT  6   // error decrypting .dat file
#define ERR_UNTARTST 7   // error (test) untarring .tar.gz
#define ERR_UNTAR    8   // error (real) untarring .tar.gz
#define ERR_DAT      9   // missing .dat file

#define MOUNTPOINT "/home/httpd/html/backup"
#define BACKUP_KEY CONFIG_ROOT"/backup/backup.key"

char hostname[STRING_SIZE];
char tmpdir[STRING_SIZE];
char tmpdatefile[STRING_SIZE];  /* where date is written */

/* check whether a file exists */
int file_exists(const char *fname)
{
	struct stat st;
	stat(fname, &st);

	if (S_ISREG(st.st_mode))
		return 1;
	else
		return 0;
}

/* Write tagfile date time without comment */
void writetagfile(const char* filename, const char* timestamp)
{
	FILE *fd;

	if(!(fd = fopen(filename,"w"))) {
		perror("ipcoprscfg : Unable to open tagfile\n");
		exit(ERR_ANY);
	}
	/* get exclusive lock to prevent a mess if 2 copies run at once */
	flock(fileno(fd), LOCK_EX);
	fprintf(fd, "%s \n", timestamp);
	flock(fileno(fd), LOCK_UN);
	fclose(fd);
}

void exithandler(void)
{
	char command[STRING_SIZE];

	/* clean up temporary files and directory */
	snprintf(command, STRING_SIZE-1,"/tmp/%s.tar.gz", hostname);
	unlink(command);
	if (tmpdir) {
		snprintf(command, STRING_SIZE-1,
			 "/bin/rm -rf %s &>/dev/null", tmpdir);
		safe_system(command);
	}
	if (tmpdatefile) unlink(tmpdatefile);
	/* remove just uploaded file */
	snprintf(command, STRING_SIZE-1,
			MOUNTPOINT "/%s.dat", hostname);
	unlink(command);
	snprintf(command, STRING_SIZE-1, "/tmp/%s.tar", hostname);
	unlink(command);
}

static void usage()
{
	fprintf(stderr,"Usage:\n");
	fprintf(stderr,"ipcoprscfg --import\n");
	fprintf(stderr,"\tRename <hostname>.dat to <hostname>.YYYY-MM-DD_HH-MM-SS.dat\n");
	fprintf(stderr,"\tand create corresponding .dat.time file\n");
	fprintf(stderr,"ipcoprscfg --restore file.dat [--hardware]\n");
	fprintf(stderr,"\tRestore a file.dat backup\n");
	fprintf(stderr,"\t--hardware option restore hardware settings\n");
	exit(1);
}

int main(int argc, char *argv[])
{
	int rshardware=0;
	int import=0;
	int tempfd;
	char command[STRING_SIZE];
	char filename[STRING_SIZE];
	char buffer[STRING_SIZE];
	char *timestamp=NULL;
	FILE *fd;

	if (argc<2) {
		/* need a minima one parameter */
		usage();
	} else if (strcmp(argv[1],"--import")==0) {
		/* After import, rename a .dat file with timestamp
		 * and create associated .dat.time */
		import=1;
	} else if (strcmp(argv[1],"--restore")==0) {
		if (argc==4 && (strcmp(argv[3],"--hardware")==0)) {
			rshardware=1; // restore hardware settings
		} else if (argc!=3) {
			fprintf(stderr, "ipcoprscfg : bad syntax \n");
			exit (ERR_ANY);
		}
	} else {
		fprintf(stderr, "ipcoprscfg : bad parameter\n");
		exit (ERR_ANY);
	}

	gethostname(hostname, STRING_SIZE-1);

	if (!(import)) {
		/* check a valid filename, full file name lenght */
		if (strlen(argv[2]) != strlen(MOUNTPOINT "/-YYYY-MM-DD_HH-MM-SS.dat")
					+strlen(hostname)) {
			fprintf( stderr, "ipcoprscfg : bad file name\n");
			exit(ERR_ANY);
		}
		/* file in the path */
		if (!strstr(argv[2], MOUNTPOINT)) {
			fprintf( stderr, "ipcoprscfg : wrong path\n");
			exit(ERR_ANY);
		}
		/* and exist */
		if (!file_exists(argv[2])) {
			fprintf (stderr, "Missing backup\n");
			exit (ERR_DAT);
		}
		strcpy(filename, argv[2]);
	} else {
		snprintf (filename, STRING_SIZE-1,
			MOUNTPOINT "/%s.dat", hostname);
	}

	/* Init setuid */
	if (!(initsetuid()))
		exit(ERR_ANY);

	/* key file and encrypted .dat are required */
	if (!file_exists(BACKUP_KEY)) {
		fprintf (stderr, "Missing encryption key\n");
		exit (ERR_DECRYPT);
	}

	/* now exithandler will have something to do */ 
	atexit(exithandler);

	/* decrypt .dat file to /tmp/hostname.tar.gz */
	snprintf (command, STRING_SIZE-1,
		"/usr/bin/openssl des3 -d -salt "
		"-in %s "
		"-out /tmp/%s.tar.gz "
		"-kfile " BACKUP_KEY
		"> /dev/null 2> /dev/null", filename, hostname);
	if (safe_system (command)) {
		fprintf (stderr, "Couldn't decrypt %s.dat archive\n", hostname);
		exit (ERR_DECRYPT);
	}

	/* create temporary directory for testing untar */
	strcpy (tmpdir, "/tmp/cfg_XXXXXX");
	if (mkdtemp (tmpdir)==NULL) {
		exit (ERR_ANY);
	}

	/* Start (test) untarring files from compressed archive */
	snprintf (command, STRING_SIZE-1, "/bin/tar -C %s -xzvf /tmp/%s.tar.gz > /dev/null 2> /dev/null", tmpdir, hostname);
	if (safe_system (command)) {
		fprintf (stderr, "Invalid archive\n");
		exit (ERR_UNTARTST);
	}

	/* If in import mode, read tagfile if include
	 * or parse backup to read newest file time change */
	if (import) {
		snprintf (filename, STRING_SIZE-1,
			"%s" CONFIG_ROOT "/backup.dat.time", tmpdir);
		if (!file_exists(filename)) {
			/* without tagfile include */
			strcpy (tmpdatefile, "/tmp/date.XXXXXX");
			if (! (tempfd = mkstemp (tmpdatefile)) > 0) {
				fprintf(stderr, "Couldn't create temporary file.\n");
				exit(ERR_ANY);
			}
			/* Parse all files of a backup 
			 * to read last modification date.
			 * The most recent date is written on tmp file */
			snprintf( command, STRING_SIZE-1,
				"/usr/bin/find %s -type f -exec "
				"/bin/ls --time-style=+%%Y-%%m-%%d_%%H-%%M-%%S -l {} \\; | "
				"/usr/bin/awk '{print $6}' | "
				"/usr/bin/sort -rn | /bin/head -n 1 >%s", tmpdir, tmpdatefile);
			if (safe_system (command)) {
				fprintf(stderr, "Couldn't write tmpdatefile\n");
				exit (ERR_ANY);
			}
			/* read the date */
			if(!(fd = fopen(tmpdatefile,"r"))) {
				perror("ipcoprscfg : Failed opening file tagfile");
				exit(ERR_ANY);
			}
			fgets(buffer, STRING_SIZE, fd);
			fclose(fd);
			if (buffer[strlen(buffer)-1] =='\n') buffer[strlen(buffer)-1] = '\0';
			timestamp=strtok(buffer," ");
			/* Write tagfile for backup */
			snprintf( filename, STRING_SIZE-1,
				MOUNTPOINT "/%s-%s.dat.time", hostname, timestamp);
			writetagfile( filename, timestamp);
		} else {
			/* date is available to be read directly */
			if(!(fd = fopen(filename,"r"))) {
				perror("ipcoprscfg : Failed opening file tagfile");
				exit(ERR_ANY);
			}
			fgets(buffer, STRING_SIZE, fd);
			fclose(fd);
			timestamp=strtok(buffer," ");

			/* Moving from ext3 to FAT system, system warn not
			 * able to preserve ownership, change before */
			chown( filename, 99, 99);
			snprintf(command, STRING_SIZE-1,
				"/bin/mv %s " MOUNTPOINT "/%s-%s.dat.time",
				filename, hostname, timestamp);
			if (safe_system(command)) {
				perror("Failed to copy tagfile");
				exit(ERR_ANY);
			}
		}
		/* backup is valid, move to a name with date */
		snprintf (filename, STRING_SIZE-1,
			MOUNTPOINT "/%s.dat", hostname);
		snprintf (command, STRING_SIZE-1,
			"/bin/mv %s " MOUNTPOINT "/%s-%s.dat",
			filename, hostname, timestamp);
		if (safe_system(command)) {
			perror("Failed to copy backup");
			exit(ERR_ANY);
		}
		exit(0);
	}

	/* uncompress archive */
	snprintf(command, STRING_SIZE-1, "/bin/gunzip -d -f /tmp/%s.tar.gz", hostname);
	safe_system (command);

	/* remove hardware specific settings */
	if (rshardware==0) {
		snprintf (command, STRING_SIZE-1,
			"/bin/tar --delete --file=/tmp/%s.tar "
			"-T " CONFIG_ROOT "/backup/exclude.hardware",
			hostname);
		safe_system (command);
	}

	/* Start (real) untarring files from compressed archive */
	snprintf (command, STRING_SIZE-1,
		"/bin/tar -X " CONFIG_ROOT "/backup/exclude.system"
		" -C / -xvf /tmp/%s.tar &>/dev/null", hostname);
	if (safe_system (command)) {
		fprintf (stderr, "Error restoring archive\n");
		exit (ERR_UNTAR);
	} else {
		if (safe_system ("/usr/local/bin/upgrade &>/dev/null")) {
			fprintf (stderr, "Error upgrading data from backup!\n");
			exit (ERR_ANY);
		}
	}
	exit(0);
}
